/*global PouchDB: true, couchEval: true */

"use strict";

function Validation(db) {
  function doValidation(db, newDoc, options, callback) {
    if ((newDoc._id || "").indexOf("_design/") === 0) {
      //a design document -> always validates succesful.
      callback(null, true)
      return;
    }
    fillInValidationOptions(db, options, function (newOptions) {
      getValidationFunctions(db, function (err, validationFuncs) {
        if (err) {
          callback(err, null);
          return;
        }
        if (!validationFuncs.length) {
          //no validation functions -> success
          callback(null, true);
          return;
        }
        getOldDoc(db, newDoc._id, function (err, oldDoc) {
          try {
            validationFuncs.forEach(function (validationFunc) {
              validationFunc(newDoc, oldDoc, newOptions.userCtx, newOptions.secObj);
            });
          } catch (e) {
            if (typeof e.unauthorized !== "undefined") {
              callback({
                error: "unauthorized",
                reason: e.unauthorized,
                status: 401
              }, null);
            } else if (typeof e.forbidden !== "undefined") {
              callback({
                error: "forbidden",
                reason: e.forbidden,
                status: 403
              }, null);
            } else {
              callback({
                error: e.name,
                reason: e.toString() + "\n\n" + e.stack,
                status: 500
              }, null);
            }
            return;
          }
          //passed all validation functions -> success
          callback(null, true);
        });
      });
    });
  }
  function fillInValidationOptions(db, options, callback) {
    if (!options) {
      options = {};
    }
    if (!options.secObj) {
      //a default security object
      options.secObj = {
        admins: {
          names: [],
          roles: []
        },
        members: {
          names: [],
          roles: []
        }
      };
    }
    if (!options.userCtx) {
      db.info(function (err, resp) {
        //a default userCtx
        options.userCtx = {
          db: resp.db_name,
          name: null,
          roles: ["_admin"]
        };
        callback(options);
      });
    } else {
      callback(options);
    }
  }

  function getValidationFunctions(db, callback) {
    db.allDocs({
      startkey: "_design/",
      endkey: "_design0",
      include_docs: true
    }, function (err, resp) {
      if (err) {
        callback(err, null);
        return;
      }
      var validationFuncs = resp.rows.map(function (row) {
        return {
          doc: row.doc,
          func: row.doc.validate_doc_update
        };
      });
      validationFuncs = validationFuncs.filter(function (info) {
        return typeof info.func !== "undefined";
      });
      var currentInfo;
      try {
        validationFuncs = validationFuncs.map(function (info) {
          currentInfo = info;
          //convert str -> function
          var func = couchEval(info.doc, {}, info.func);
          if (typeof func !== "function") {
            //activate the exception handling mechanism down here.
            throw new Error();
          }
          return func;
        });
      } catch (e) {
        callback({
          "error": "compilation_error",
          "status": 500,
          "reason": "Expression does not eval to a function. " + currentInfo.func
        }, null);
        return;
      }
      callback(null, validationFuncs);
    });
  }

  function getOldDoc(db, id, callback) {
    if (!id) {
      //error message for internal use only.
      callback({"error": "Invalid id..."}, null);
    }
    db.get(id, function (err, resp) {
      if (err) {
        callback(err, null);
      } else {
        callback(null, resp);
      }
    });
  }

  function processArgs(callback, options) {
    if (!callback) {
      callback = options;
      options = {};
    }
    return {
      callback: callback,
      options: options
    };
  }

  return {
    validating: {
      put: function (doc, options, callback) {
        var args = processArgs(callback, options);
        doValidation(db, doc, args.options, function (err, success) {
          if (err) {
            args.callback(err, null);
            return;
          }
          db.put(doc, args.options, args.callback);
        });
      },
      post: function (doc, options, callback) {
        var args = processArgs(callback, options);
        doValidation(db, doc, args.options, function (err, success) {
          if (err) {
            args.callback(err, null);
            return;
          }
          db.post(doc, args.options, args.callback);
        });
      },
      remove: function (doc, options, callback) {
        var args = processArgs(callback, options);

        doc._deleted = true;
        doValidation(db, doc, args.options, function (err, success) {
          if (err) {
            args.callback(err, null);
            return;
          }
          db.remove(doc, args.options, args.callback);
        });
      },
      bulkDocs: function (bulkDocs, options, callback) {
        var args = processArgs(callback, options);

        var passes = [];
        var failes = [];
        bulkDocs.docs.forEach(function (doc) {
          doValidation(db, doc, args.options, function (err, success) {
            if (err) {
              failes.push(err);
            } else {
              passes.push(success);
            }
            if (passes.length + failes.length === bulkDocs.docs.length) {
              if (failes.length) {
                args.callback(failes, null);
                return;
              }
              db.bulkDocs(bulkDocs, args.options, args.callback);
            }
          });
        });
      },
      putAttachment: function (docId, attachmentId, rev, attachment, type, options, callback) {
        var args = processArgs(callback, options);
        db.get(docId, {rev: rev}, function (err, doc) {
          if (err) {
            doc = {_id: docId};
          }
          doc._attachments = doc._attachments || {};
          doc._attachments[attachmentId] = {
            content_type: type,
            data: attachment,
          };
          doValidation(db, doc, args.options, function (err, success) {
            if (err) {
              args.callback(err, null);
              return;
            }
            db.putAttachment(docId, attachmentId, rev, attachment, type, args.callback);
          });
        });
      },
      removeAttachment: function (docId, attachmentId, rev, options, callback) {
        var args = processArgs(callback, options);
        db.get(docId, {rev: rev}, function (err, doc) {
          delete doc._attachments[attachmentId];

          doValidation(db, doc, args.options, function (err, success) {
            if (err) {
              args.callback(err, null);
              return;
            }
            db.removeAttachment(docId, attachmentId, rev, args.callback);
          });
        });
      }
    }
  };
}
//no need to track deletes
Validation._delete = function () {};

PouchDB.plugin('validation', Validation);
